import torch
import os
import cv2
import albumentations as A
from albumentations.pytorch import ToTensorV2
from torch.utils.data import DataLoader
from tqdm import tqdm


class SegDataset(torch.utils.data.Dataset):
    def __init__(self, images_dir, masks_dir):
        self.transforms = A.Compose([
            A.Resize(512, 512),
            A.Normalize(),
            ToTensorV2()
        ])
        self.ids = os.listdir(images_dir)
        self.images_fps = [os.path.join(images_dir, image_id) for image_id in self.ids]
        self.masks_fps = [os.path.join(masks_dir, image_id.replace(".jpg", ".png")) for image_id in self.ids]

    def __getitem__(self, i):
        image = cv2.imread(self.images_fps[i])
        mask = cv2.imread(self.masks_fps[i])
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        mask = cv2.cvtColor(mask, cv2.COLOR_BGR2RGB)
        transformed = self.transforms(image=image, mask=mask)
        return transformed['image'], transformed['mask'][:, :, 0], self.ids[i]  # 返回文件名

    def __len__(self):
        return len(self.ids)


def inference_and_save(model, data_loader, device, save_dir):
    model.eval()
    os.makedirs(save_dir, exist_ok=True)

    with torch.no_grad():
        for images, _, filenames in tqdm(data_loader, total=len(data_loader), desc="Inference"):
            images = images.to(device)
            predictions = model(images)
            predictions = predictions.argmax(dim=1).cpu().numpy()

            for idx, (pred, filename) in enumerate(zip(predictions, filenames)):
                # 将jpg扩展名改为png
                output_filename = os.path.splitext(filename)[0] + ".png"
                result_path = os.path.join(save_dir, output_filename)
                cv2.imwrite(result_path, pred)


x_test_dir = "./data/test/images"
y_test_dir = "./data/test/annotations"
test_dataset = SegDataset(x_test_dir, y_test_dir)
test_loader = DataLoader(test_dataset, batch_size=1, shuffle=False, drop_last=False)
device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
model = torch.load("./checkpoints/model.pth")
save_dir = "./infer_results"
inference_and_save(model, test_loader, device, save_dir)
print(f"Results saved in {save_dir}")
