import numpy as np
from pathlib import Path
from tqdm import tqdm


def calculate_iou_metrics(gt_folder, pred_folder, num_classes=3):
    # 创建混淆矩阵
    confusion_matrix = np.zeros((num_classes, num_classes), dtype=np.int64)

    # 获取所有ground truth文件
    gt_files = list(Path(gt_folder).glob("*.npy"))

    if not gt_files:
        print(f"在{gt_folder}中没有找到npy格式的ground truth文件")
        return

    print(f"找到 {len(gt_files)} 个ground truth文件")

    # 处理每个文件
    for gt_file in tqdm(gt_files, desc="计算IoU"):
        # 获取对应的预测文件名
        filename = gt_file.stem
        pred_file = Path(pred_folder) / f"{filename}.npy"

        # 检查预测文件是否存在
        if not pred_file.exists():
            print(f"预测文件不存在: {pred_file}")
            continue

        # 读取ground truth和预测结果
        gt = np.load(str(gt_file))
        pred = np.load(str(pred_file))

        # 确保两个图像具有相同的形状
        if gt.shape != pred.shape:
            print(f"形状不匹配: gt={gt.shape}, pred={pred.shape}, 文件={filename}")
            # 可以选择调整大小或跳过这个文件
            continue

        # 更新混淆矩阵
        for i in range(num_classes):
            for j in range(num_classes):
                confusion_matrix[i, j] += np.sum((gt == i) & (pred == j))

    # 计算每个类别的IoU
    iou_per_class = np.zeros(num_classes)
    for i in range(num_classes):
        # IoU = 真阳性 / (真阳性 + 假阳性 + 假阴性)
        intersection = confusion_matrix[i, i]
        union = (np.sum(confusion_matrix[i, :]) +
                 np.sum(confusion_matrix[:, i]) -
                 confusion_matrix[i, i])

        if union > 0:
            iou_per_class[i] = intersection / union
        else:
            iou_per_class[i] = 0.0

    # 计算mIoU
    miou = np.mean(iou_per_class)

    return iou_per_class, miou, confusion_matrix


if __name__ == "__main__":
    import time

    start_time = time.time()

    # 文件夹路径
    gt_folder = "./ground_truth"  # 存放ground truth的文件夹(npy格式)
    pred_folder = "./infer_results_npy"  # 存放推理结果的文件夹(npy格式)

    # 计算IoU和mIoU
    iou_per_class, miou, conf_matrix = calculate_iou_metrics(
        gt_folder, pred_folder,
        num_classes=3,  # 假设有3个类别(0,1,2)
    )

    # 计算总耗时
    total_time = time.time() - start_time

    # 直接打印结果，不保存到文件
    print("\n===== 最终评估结果 =====")
    print("类别IoU结果:")
    for i in range(len(iou_per_class)):
        print(f"类别 {i}: IoU = {iou_per_class[i]:.4f}")
    print(f"\n总体 mIoU = {miou:.4f}")
    print(f"评估总耗时: {total_time:.2f}秒")
