from tqdm import tqdm
from torch.utils.data import DataLoader
import numpy as np
import os
import cv2
import albumentations as A
from albumentations.pytorch import ToTensorV2
import torch
import torch.nn as nn
from time import time
from model import Unet


class SegDataset(torch.utils.data.Dataset):
    def __init__(self, images_dir, masks_dir):
        # 使用Albumentations定义变换
        self.transforms = A.Compose([
            A.Resize(512, 512),  # 调整图像大小
            A.Normalize(),  # 归一化
            ToTensorV2()  # 转换为张量
        ])

        self.ids = os.listdir(images_dir)  # 单个目录直接使用
        self.images_fps = [images_dir + "/" + image_id for image_id in self.ids]
        self.masks_fps = [masks_dir + "/" + image_id for image_id in self.ids]

    def __getitem__(self, i):
        # 使用cv2读取图像
        image = cv2.imread(self.images_fps[i])
        mask = cv2.imread(self.masks_fps[i].replace(".jpg", ".png"))

        # 将图像和掩码从BGR转换为RGB
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        mask = cv2.cvtColor(mask, cv2.COLOR_BGR2GRAY)

        image = np.array(image)
        mask = np.array(mask)

        # 应用预处理变换
        transformed = self.transforms(image=image, mask=mask)

        # 返回变换后的图像和掩码（假设掩码是单通道的，选择第一个通道）
        return transformed['image'], transformed['mask']

    def __len__(self):
        return len(self.ids)


def evaluate_accuracy_gpu(net, data_iter, device=None):
    if device is None and isinstance(net, nn.Module):
        device = next(iter(net.parameters())).device
    net.eval()
    metric = [0.0, 0.0]
    with torch.no_grad():
        for X, y in data_iter:
            if isinstance(X, list):
                X = [x.to(device) for x in X]
            else:
                X = X.to(device)
            y = y.to(device)
            metric[0] += (net(X).argmax(dim=1) == y).float().sum().item()
            metric[1] += y.numel()
    return metric[0] / metric[1]


def train_batch(net, features, labels, loss, trainer, device):
    if isinstance(features, list):
        features = [x.to(device) for x in features]
    else:
        features = features.to(device)
    labels = labels.to(device)
    net.train()
    trainer.zero_grad()
    predictions = net(features)
    l = loss(predictions, labels)
    l.backward()
    trainer.step()
    with torch.no_grad():
        acc = (predictions.argmax(dim=1) == labels).float().mean().item()
    return l.item(), acc


def train(net, train_iter, test_iter, loss, trainer, num_epochs, scheduler, device=None):
    net = net.to(device)
    model_save_path = "checkpoints/model_test.pth"
    dir_name = os.path.dirname(model_save_path)
    if not os.path.exists(dir_name):
        os.makedirs(dir_name)

    for epoch in range(num_epochs):
        metric = [0.0, 0.0, 0.0, 0.0]  # 训练损失总和，训练准确率总和，样本数，总预测数
        start_time = time()

        for i, (features, labels) in tqdm(enumerate(train_iter), total=len(train_iter),
                                          desc=f"Training Epoch [{epoch + 1}/{num_epochs}]"):
            l, acc = train_batch(net, features, labels.long(), loss, trainer, device)
            metric[0] += l * labels.shape[0]
            metric[1] += acc * labels.numel()
            metric[2] += labels.shape[0]
            metric[3] += labels.numel()

        scheduler.step()
        test_acc = evaluate_accuracy_gpu(net, test_iter, device)
        epoch_time = time() - start_time
        print(
            f"Epoch {epoch + 1} \n--- Loss {metric[0] / metric[2]:.3f} \n--- Train Acc {metric[1] / metric[3]:.3f} \n"
            f"--- Test Acc {test_acc:.3f} \n--- Time {epoch_time:.1f}s")

    # 保存最终模型
    torch.save(net, model_save_path)
    print(f"Final model saved at {model_save_path}")


x_train_dir = "./data/train/images"
y_train_dir = "./data/train/annotations"
x_test_dir = "./data/test/images"
y_test_dir = "./data/test/annotations"

train_dataset = SegDataset(
    x_train_dir,
    y_train_dir,
)
test_dataset = SegDataset(
    x_test_dir,
    y_test_dir,
)

batch_size = 8
num_epochs = 1

train_loader = DataLoader(train_dataset, batch_size=batch_size, shuffle=True, drop_last=True)
test_loader = DataLoader(test_dataset, batch_size=batch_size, shuffle=True, drop_last=True)

device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
model = Unet(num_classes=3).to(device)
lossf = nn.CrossEntropyLoss(ignore_index=255)

optimizer = torch.optim.AdamW(model.parameters(), lr=0.01, weight_decay=1e-8)
scheduler = torch.optim.lr_scheduler.StepLR(optimizer, step_size=100, gamma=0.1, last_epoch=-1)
train(model, train_loader, test_loader, lossf, optimizer, num_epochs, scheduler, device=device)
